/* GTK+ DirectFB Engine
 * Copyright (C) 2001 convergence integrated media GmbH
 * Copyright (C) 1998-2000 Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Written by Sven Neumann <sven@convergence.de>, 
 * based on code by Owen Taylor <otaylor@redhat.com> 
 * and Carsten Haitzler <raster@rasterman.com>.
 */

#include <stdlib.h>

#include <gdkprivate-directfb.h>

#include "directfb-engine.h"


/*  eeek, this is defined in gdkregion-generic.h  */

typedef GdkSegment GdkRegionBox;
struct _GdkRegion
{
  long          size;
  long          numRects;
  GdkRegionBox *rects;
  GdkRegionBox  extents;
};


static IDirectFBSurface * surface_cache_value_new (gchar            *filename);
static void               surface_destroy         (IDirectFBSurface *surface);
static void               surface_render          (IDirectFBSurface *src,
						   GdkColor         *color,
						   GdkWindow        *window,
						   GdkRectangle     *clip_rect,
						   gint              src_x,
						   gint              src_y,
						   gint              src_width,
						   gint              src_height,
						   gint              dest_x,
						   gint              dest_y,
						   gint              dest_width,
						   gint              dest_height,
                                                   gboolean          stretch);


static GCache *surface_cache = NULL;


static IDirectFBSurface *
surface_cache_value_new (gchar *filename)
{
  IDirectFBSurface       *surface;
  DFBSurfaceDescription   dsc;
  IDirectFBImageProvider *provider;
  
  g_return_val_if_fail (filename != NULL, NULL);

  if (DirectFB->CreateImageProvider (DirectFB, filename, &provider) != DFB_OK)
    {
      g_warning ("DirectFB theme: Cannot load image for pixmap %s\n", 
		 filename);
      return NULL;
    }

  provider->GetSurfaceDescription (provider, &dsc);

  if (DirectFB->CreateSurface (DirectFB, &dsc, &surface) != DFB_OK)
    {
      g_warning ("DirectFB theme: Cannot create surface for pixmap %s\n", 
		 filename);
      provider->Release (provider);
      free (provider);
      return NULL;
    }

  provider->RenderTo (provider, surface);
    
  provider->Release (provider);
  free (provider);

  return surface;
}

static void
surface_destroy (IDirectFBSurface *surface)
{
  surface->Release (surface);
  free (surface);
}

static void
surface_render (IDirectFBSurface *src,
		GdkColor         *color,
		GdkDrawable      *destination,
		GdkRectangle     *clip_rect,
		gint              src_x,
		gint              src_y,
		gint              src_width,
		gint              src_height,
		gint              dest_x,
		gint              dest_y,
		gint              dest_width,
		gint              dest_height,
                gboolean          stretch)
{
  GdkWindowObject         *private = NULL;
  GdkDrawableImplDirectFB *impl;
  IDirectFBSurface        *dest;
  DFBRectangle             src_rect;
  DFBRectangle             dest_rect;
  GList                   *list;
  gint                     i;
  gint                     x, y;
  GdkRegion               *clip_region;
  GdkRegion               *tmp_region;
  GdkRectangle             rect = { dest_x, dest_y, 
				    dest_width, dest_height };

  if (src_width <= 0 || src_height <= 0)
    return;

  if (dest_width <= 0 || dest_height <= 0)
    return;

  if (GDK_IS_PIXMAP (destination))
    {
      impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_PIXMAP_OBJECT (destination)->impl);
    }
  else if (GDK_IS_WINDOW (destination))
    {
      private = GDK_WINDOW_OBJECT (destination);
      impl = GDK_DRAWABLE_IMPL_DIRECTFB (private->impl);
    }
  else if (GDK_IS_DRAWABLE_IMPL_DIRECTFB (destination))
    {
      impl = GDK_DRAWABLE_IMPL_DIRECTFB (destination);
    }
  else
    return;

  dest = impl->surface;

  if (FALSE && color)  /*  temporarily disabled  */
    {
      dest->SetColor (dest, 
		      color->red   >> 8, 
		      color->green >> 8, 
		      color->blue  >> 8, 0xFF);
      dest->SetBlittingFlags (dest, (DSBLIT_BLEND_ALPHACHANNEL | 
				     DSBLIT_COLORIZE));
    }
  else
    {
      dest->SetBlittingFlags (dest, DSBLIT_BLEND_ALPHACHANNEL);
    }

  src_rect.x = src_x;
  src_rect.y = src_y;
  src_rect.w = src_width;
  src_rect.h = src_height;
  
  dest_rect.x = dest_x;
  dest_rect.y = dest_y;
  dest_rect.w = dest_width;
  dest_rect.h = dest_height;

  if (clip_rect)
    gdk_rectangle_intersect (&rect, clip_rect, &rect);

  clip_region = gdk_region_rectangle (&rect);

  if (impl->buffered && impl->paint_region)
    gdk_region_intersect (clip_region, impl->paint_region);

  if (private)
    for (list = private->children; list; list = list->next)
      {
	GdkDrawableImplDirectFB *child_impl;
	GdkWindowObject         *child = GDK_WINDOW_OBJECT (list->data);

	if (child->state & GDK_WINDOW_STATE_WITHDRAWN || child->input_only)
	  continue;
    
	child_impl = GDK_DRAWABLE_IMPL_DIRECTFB (child->impl);

	rect.x      = child->x; 
	rect.y      = child->y;
	rect.width  = child_impl->width;
	rect.height = child_impl->height;

	tmp_region = gdk_region_rectangle (&rect);
	gdk_region_subtract (clip_region, tmp_region); 
	gdk_region_destroy (tmp_region);
      }
  
  for (i = 0; i < clip_region->numRects; i++)
    {
      /*  should probably blit the correct regions instead of using clip  */
      dest->SetClip (dest, (DFBRegion *) &clip_region->rects[i]);

      if (dest_width == src_width && dest_height == src_height)
        {
          dest->Blit (dest, src, &src_rect, dest_x, dest_y);
        }
      else if (stretch)
        {
	  dest->StretchBlit (dest, src, &src_rect, &dest_rect);

	  /*
	  g_print ("StretchBlit: %d,%d (%dx%d)  ->  %d,%d (%dx%d)\n",
		   src_rect.x, src_rect.y, src_rect.w, src_rect.h,
		   dest_rect.x, dest_rect.y, dest_rect.w, dest_rect.h);
	  */
        }
      else
        {
          for (y = 0; y < dest_height; y += src_height)
            for (x = 0; x < dest_width; x += src_width)
              dest->Blit (dest, src, &src_rect, dest_x + x, dest_y + y);
        }

      gdk_directfb_add_update (GDK_DRAWABLE_IMPL_DIRECTFB (impl), 
			       (DFBRegion *) &clip_region->rects[i]);
    }

  dest->SetClip (dest, NULL);
  dest->SetBlittingFlags (dest, 0);
}

ThemeSurface *
theme_surface_new (void)
{
  ThemeSurface *result;

  result = g_new (ThemeSurface, 1);

  result->filename      = NULL;
  result->surface       = NULL;

  result->stretch       = TRUE;
  result->border_left   = 0;
  result->border_right  = 0;
  result->border_bottom = 0;
  result->border_top    = 0;

  return result;
}

void
theme_surface_destroy (ThemeSurface *theme_surface)
{
  if (theme_surface->surface)
    g_cache_remove (surface_cache, theme_surface->surface);
}

void      
theme_surface_set_filename (ThemeSurface *theme_surface,
			    const gchar  *filename)
{
  if (theme_surface->surface)
    {
      g_cache_remove (surface_cache, theme_surface->surface);
      theme_surface->surface = NULL;
    }

  if (theme_surface->filename)
    g_free (theme_surface->filename);

  theme_surface->filename = g_strdup (filename);
}

void
theme_surface_set_border (ThemeSurface *theme_surface,
			  gint          left,
			  gint          right,
			  gint          top,
			  gint          bottom)
{
  theme_surface->border_left   = left;
  theme_surface->border_right  = right;
  theme_surface->border_top    = top;
  theme_surface->border_bottom = bottom;
}

void
theme_surface_set_stretch (ThemeSurface *theme_surface,
			   gboolean      stretch)
{
  theme_surface->stretch = stretch;
}

IDirectFBSurface *
theme_surface_get_surface (ThemeSurface *theme_surface)
{
  if (!theme_surface->filename)
    return NULL;

  if (!theme_surface->surface)
    {
      if (!surface_cache)
	surface_cache = g_cache_new ((GCacheNewFunc) surface_cache_value_new,
				    (GCacheDestroyFunc) surface_destroy,
				    (GCacheDupFunc) g_strdup,
				    (GCacheDestroyFunc) g_free,
				     g_str_hash, g_direct_hash, g_str_equal);
      
      theme_surface->surface = g_cache_insert (surface_cache, 
					       theme_surface->filename);
    }
  
  return theme_surface->surface;
}

void
theme_surface_render (ThemeSurface *theme_surface,
		      GdkColor     *color,
		      GdkWindow    *window,
		      GdkRectangle *clip_rect,
		      guint         component_mask,
		      gboolean      center,
		      gint          x,
		      gint          y,
		      gint          width,
		      gint          height)
{
  IDirectFBSurface *surface;
  gint src_x[4], src_y[4], dest_x[4], dest_y[4];
  gint surface_width;
  gint surface_height;

  surface = theme_surface_get_surface (theme_surface);

  if (!surface)
    return;

  surface->GetSize (surface, &surface_width, &surface_height);

  if (center && !theme_surface->stretch)
    {
      x += (width - surface_width) / 2;
      y += (height - surface_height) / 2;
      
      surface_render (surface, color, window, clip_rect,
                      0, 0,
                      surface_width, surface_height,
                      x, y,
                      surface_width, surface_height,
                      FALSE);
    }
  else
    {
      src_x[0] = 0;
      src_x[1] = theme_surface->border_left;
      src_x[2] = surface_width - theme_surface->border_right;
      src_x[3] = surface_width;
      
      src_y[0] = 0;
      src_y[1] = theme_surface->border_top;
      src_y[2] = surface_height - theme_surface->border_bottom;
      src_y[3] = surface_height;
      
      dest_x[0] = x;
      dest_x[1] = x + theme_surface->border_left;
      dest_x[2] = x + width - theme_surface->border_right;
      dest_x[3] = x + width;

      dest_y[0] = y;
      dest_y[1] = y + theme_surface->border_top;
      dest_y[2] = y + height - theme_surface->border_bottom;
      dest_y[3] = y + height;

      if (component_mask & COMPONENT_ALL)
	component_mask = (COMPONENT_ALL - 1) & ~component_mask;

#define RENDER_COMPONENT(X1,X2,Y1,Y2)                                   \
       surface_render (surface, color, window, clip_rect,               \
	 	       src_x[X1], src_y[Y1],                            \
		       src_x[X2] - src_x[X1], src_y[Y2] - src_y[Y1],    \
		       dest_x[X1], dest_y[Y1],                          \
		       dest_x[X2] - dest_x[X1], dest_y[Y2] - dest_y[Y1],\
                       theme_surface->stretch)
      
      if (component_mask & COMPONENT_NORTH_WEST)
	RENDER_COMPONENT (0, 1, 0, 1);

      if (component_mask & COMPONENT_NORTH)
	RENDER_COMPONENT (1, 2, 0, 1);

      if (component_mask & COMPONENT_NORTH_EAST)
	RENDER_COMPONENT (2, 3, 0, 1);

      if (component_mask & COMPONENT_WEST)
	RENDER_COMPONENT (0, 1, 1, 2);

      if (component_mask & COMPONENT_CENTER)
	RENDER_COMPONENT (1, 2, 1, 2);

      if (component_mask & COMPONENT_EAST)
	RENDER_COMPONENT (2, 3, 1, 2);

      if (component_mask & COMPONENT_SOUTH_WEST)
	RENDER_COMPONENT (0, 1, 2, 3);

      if (component_mask & COMPONENT_SOUTH)
	RENDER_COMPONENT (1, 2, 2, 3);

      if (component_mask & COMPONENT_SOUTH_EAST)
	RENDER_COMPONENT (2, 3, 2, 3);
    }
}
